# Test of opening and closing and opening

import fiona


def test_write_revolving_door(tmpdir, path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as src:
        meta = src.meta
        features = list(src)

    shpname = str(tmpdir.join('foo.shp'))

    with fiona.open(shpname, 'w', **meta) as dst:
        dst.writerecords(features)

    with fiona.open(shpname) as src:
        pass



"""Tests of features related to GDAL RFC 49

See https://trac.osgeo.org/gdal/wiki/rfc49_curve_geometries.
"""

import fiona

from .conftest import requires_gdal2


@requires_gdal2
def test_line_curve_conversion(path_curves_line_csv):
    """Convert curved geometries to linear approximations"""
    with fiona.open(path_curves_line_csv) as col:
        assert col.schema['geometry'] == 'Unknown'
        features = list(col)
        assert len(features) == 9



"""Tests of fiona.env"""

import pytest

import fiona
import fiona.env
from fiona.session import AWSSession


def test_nested_credentials(monkeypatch):
    """Check that rasterio.open() doesn't wipe out surrounding credentials"""

    @fiona.env.ensure_env_with_credentials
    def fake_opener(path):
        return fiona.env.getenv()

    with fiona.env.Env(session=AWSSession(aws_access_key_id='foo', aws_secret_access_key='bar')):
        assert fiona.env.getenv()['AWS_ACCESS_KEY_ID'] == 'foo'
        assert fiona.env.getenv()['AWS_SECRET_ACCESS_KEY'] == 'bar'

        monkeypatch.setenv('AWS_ACCESS_KEY_ID', 'lol')
        monkeypatch.setenv('AWS_SECRET_ACCESS_KEY', 'wut')
        gdalenv = fake_opener('s3://foo/bar')
        assert gdalenv['AWS_ACCESS_KEY_ID'] == 'foo'
        assert gdalenv['AWS_SECRET_ACCESS_KEY'] == 'bar'



"""
See also test_rfc3339.py for datetime parser tests.
"""

import fiona
import pytest
import tempfile, shutil
import os
from fiona.errors import DriverSupportError
from .conftest import requires_gpkg

GDAL_MAJOR_VER = fiona.get_gdal_version_num() // 1000000

GEOMETRY_TYPE = "Point"
GEOMETRY_EXAMPLE = {"type": "Point", "coordinates": [1, 2]}

DRIVER_FILENAME = {
    "ESRI Shapefile": "test.shp",
    "GPKG": "test.gpkg",
    "GeoJSON": "test.geojson",
    "MapInfo File": "test.tab",
}

DATE_EXAMPLE = "2018-03-25"
DATETIME_EXAMPLE = "2018-03-25T22:49:05"
TIME_EXAMPLE = "22:49:05"

class TestDateFieldSupport:
    def write_data(self, driver):
        filename = DRIVER_FILENAME[driver]
        temp_dir = tempfile.mkdtemp()
        path = os.path.join(temp_dir, filename)
        schema = {
            "geometry": GEOMETRY_TYPE,
            "properties": {
                "date": "date",
            }
        }
        records = [
            {
                "geometry": GEOMETRY_EXAMPLE,
                "properties": {
                    "date": DATE_EXAMPLE,
                }
            },
            {
                "geometry": GEOMETRY_EXAMPLE,
                "properties": {
                    "date": None,
                }
            },
        ]
        with fiona.open(path, "w", driver=driver, schema=schema) as collection:
            collection.writerecords(records)
            
        with fiona.open(path, "r") as collection:
            schema = collection.schema
            features = list(collection)

        shutil.rmtree(temp_dir)

        return schema, features

    def test_shapefile(self):
        driver = "ESRI Shapefile"
        schema, features = self.write_data(driver)
        
        assert schema["properties"]["date"] == "date"
        assert features[0]["properties"]["date"] == DATE_EXAMPLE
        assert features[1]["properties"]["date"] is None

    @requires_gpkg
    def test_gpkg(self):
        driver = "GPKG"
        schema, features = self.write_data(driver)
        
        assert schema["properties"]["date"] == "date"
        assert features[0]["properties"]["date"] == DATE_EXAMPLE
        assert features[1]["properties"]["date"] is None

    def test_geojson(self):
        # GDAL 1: date field silently converted to string
        # GDAL 1: date string format uses / instead of -
        driver = "GeoJSON"
        schema, features = self.write_data(driver)
        
        if GDAL_MAJOR_VER >= 2:
            assert schema["properties"]["date"] == "date"
            assert features[0]["properties"]["date"] == DATE_EXAMPLE
        else:
            assert schema["properties"]["date"] == "str"
            assert features[0]["properties"]["date"] == "2018/03/25"
        assert features[1]["properties"]["date"] is None

    def test_mapinfo(self):
        driver = "MapInfo File"
        schema, features = self.write_data(driver)
        
        assert schema["properties"]["date"] == "date"
        assert features[0]["properties"]["date"] == DATE_EXAMPLE
        assert features[1]["properties"]["date"] is None


class TestDatetimeFieldSupport:
    def write_data(self, driver):
        filename = DRIVER_FILENAME[driver]
        temp_dir = tempfile.mkdtemp()
        path = os.path.join(temp_dir, filename)
        schema = {
            "geometry": GEOMETRY_TYPE,
            "properties": {
                "datetime": "datetime",
            }
        }
        records = [
            {
                "geometry": GEOMETRY_EXAMPLE,
                "properties": {
                    "datetime": DATETIME_EXAMPLE,
                }
            },
            {
                "geometry": GEOMETRY_EXAMPLE,
                "properties": {
                    "datetime": None,
                }
            },
        ]
        with fiona.open(path, "w", driver=driver, schema=schema) as collection:
            collection.writerecords(records)
            
        with fiona.open(path, "r") as collection:
            schema = collection.schema
            features = list(collection)

        shutil.rmtree(temp_dir)

        return schema, features

    def test_shapefile(self):
        # datetime is silently converted to date
        driver = "ESRI Shapefile"

        with pytest.raises(DriverSupportError):
            schema, features = self.write_data(driver)
        
        # assert schema["properties"]["datetime"] == "date"
        # assert features[0]["properties"]["datetime"] == "2018-03-25"
        # assert features[1]["properties"]["datetime"] is None

    @requires_gpkg
    def test_gpkg(self):
        # GDAL 1: datetime silently downgraded to date
        driver = "GPKG"
        
        if GDAL_MAJOR_VER >= 2:
            schema, features = self.write_data(driver)
            assert schema["properties"]["datetime"] == "datetime"
            assert features[0]["properties"]["datetime"] == DATETIME_EXAMPLE
            assert features[1]["properties"]["datetime"] is None
        else:
            with pytest.raises(DriverSupportError):
                schema, features = self.write_data(driver)

    def test_geojson(self):
        # GDAL 1: datetime silently converted to string
        # GDAL 1: date string format uses / instead of -
        driver = "GeoJSON"
        schema, features = self.write_data(driver)
        
        if GDAL_MAJOR_VER >= 2:
            assert schema["properties"]["datetime"] == "datetime"
            assert features[0]["properties"]["datetime"] == DATETIME_EXAMPLE
        else:
            assert schema["properties"]["datetime"] == "str"
            assert features[0]["properties"]["datetime"] == "2018/03/25 22:49:05"
        assert features[1]["properties"]["datetime"] is None

    def test_mapinfo(self):
        driver = "MapInfo File"
        schema, features = self.write_data(driver)
        
        assert schema["properties"]["datetime"] == "datetime"
        assert features[0]["properties"]["datetime"] == DATETIME_EXAMPLE
        assert features[1]["properties"]["datetime"] is None


class TestTimeFieldSupport:
    def write_data(self, driver):
        filename = DRIVER_FILENAME[driver]
        temp_dir = tempfile.mkdtemp()
        path = os.path.join(temp_dir, filename)
        schema = {
            "geometry": GEOMETRY_TYPE,
            "properties": {
                "time": "time",
            }
        }
        records = [
            {
                "geometry": GEOMETRY_EXAMPLE,
                "properties": {
                    "time": TIME_EXAMPLE,
                }
            },
            {
                "geometry": GEOMETRY_EXAMPLE,
                "properties": {
                    "time": None,
                }
            },
        ]
        with fiona.open(path, "w", driver=driver, schema=schema) as collection:
            collection.writerecords(records)
            
        with fiona.open(path, "r") as collection:
            schema = collection.schema
            features = list(collection)

        shutil.rmtree(temp_dir)

        return schema, features

    def test_shapefile(self):
        # no support for time fields
        driver = "ESRI Shapefile"
        with pytest.raises(DriverSupportError):
            self.write_data(driver)

    @requires_gpkg
    def test_gpkg(self):
        # GDAL 2: time field is silently converted to string
        # GDAL 1: time field dropped completely
        driver = "GPKG"

        with pytest.raises(DriverSupportError):
            schema, features = self.write_data(driver)
    
        # if GDAL_MAJOR_VER >= 2:
        #     assert schema["properties"]["time"] == "str"
        #     assert features[0]["properties"]["time"] == TIME_EXAMPLE
        #     assert features[1]["properties"]["time"] is None
        # else:
        #     assert "time" not in schema["properties"]

    def test_geojson(self):
        # GDAL 1: time field silently converted to string
        driver = "GeoJSON"
        schema, features = self.write_data(driver)
    
        if GDAL_MAJOR_VER >= 2:
            assert schema["properties"]["time"] == "time"
        else:
            assert schema["properties"]["time"] == "str"
        assert features[0]["properties"]["time"] == TIME_EXAMPLE
        assert features[1]["properties"]["time"] is None

    def test_mapinfo(self):
        # GDAL 2: null time is converted to 00:00:00 (regression?)
        driver = "MapInfo File"
        schema, features = self.write_data(driver)
    
        assert schema["properties"]["time"] == "time"
        assert features[0]["properties"]["time"] == TIME_EXAMPLE
        if GDAL_MAJOR_VER >= 2:
            assert features[1]["properties"]["time"] == "00:00:00"
        else:
            assert features[1]["properties"]["time"] is None



from __future__ import division
import json
import logging

import click
from cligj import use_rs_opt

from .helpers import obj_gen, eval_feature_expression
from fiona.fio import with_context_env


@click.command(short_help="Calculate GeoJSON property by Python expression")
@click.argument('property_name')
@click.argument('expression')
@click.option('--overwrite', is_flag=True, default=False,
              help="Overwrite properties, default: False")
@use_rs_opt
@click.pass_context
@with_context_env
def calc(ctx, property_name, expression, overwrite, use_rs):
    """
    Create a new property on GeoJSON features using the specified expression.

    \b
    The expression is evaluated in a restricted namespace containing:
        - sum, pow, min, max and the imported math module
        - shape (optional, imported from shapely.geometry if available)
        - bool, int, str, len, float type conversions
        - f (the feature to be evaluated,
             allows item access via javascript-style dot notation using munch)

    The expression will be evaluated for each feature and its
    return value will be added to the properties
    as the specified property_name. Existing properties will not
    be overwritten by default (an Exception is raised).

    Example

    \b
    $ fio cat data.shp | fio calc sumAB  "f.properties.A + f.properties.B"
    """
    logger = logging.getLogger(__name__)
    stdin = click.get_text_stream('stdin')
    try:
        source = obj_gen(stdin)
        for i, obj in enumerate(source):
            features = obj.get('features') or [obj]
            for j, feat in enumerate(features):

                if not overwrite and property_name in feat['properties']:
                    raise click.UsageError(
                        '{0} already exists in properties; '
                        'rename or use --overwrite'.format(property_name))

                feat['properties'][property_name] = eval_feature_expression(
                    feat, expression)

                if use_rs:
                    click.echo(u'\u001e', nl=False)
                click.echo(json.dumps(feat))

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()



# coding=utf-8
"""Encoding tests"""

from glob import glob
import os
import shutil

import pytest

import fiona


@pytest.fixture(scope='function')
def gre_shp_cp1252(tmpdir):
    """A tempdir containing copies of gre.* files, .cpg set to cp1252

    The shapefile attributes are in fact utf-8 encoded.
    """
    test_files = glob(os.path.join(os.path.dirname(__file__), 'data/gre.*'))
    tmpdir = tmpdir.mkdir('data')
    for filename in test_files:
        shutil.copy(filename, str(tmpdir))
    tmpdir.join('gre.cpg').write('cp1252')
    yield tmpdir.join('gre.shp')


def test_broken_encoding(gre_shp_cp1252):
    """Reading as cp1252 mis-encodes a Russian name"""
    with fiona.open(str(gre_shp_cp1252)) as src:
        assert next(iter(src))['properties']['name_ru'] != u'Гренада'


def test_override_encoding(gre_shp_cp1252):
    """utf-8 override succeeds"""
    with fiona.open(str(gre_shp_cp1252), encoding='utf-8') as src:
        assert next(iter(src))['properties']['name_ru'] == u'Гренада'
